/*
   This file is from bootpc 0.45 and is heavily modified for my
   Linux installation floppy and its 'setup' program.

   Florian La Roche   florian@jurix.jura.uni-sb.de


   #if 1
   Last updated : Wed Aug  2 16:42:39 1995
   Modified by JSP from code by Charles Hawkins <ceh@eng.cam.ac.uk>,

   J.S.Peatfield@damtp.cam.ac.uk

   Copyright (c) University of Cambridge, 1993,1994,1995
   #endif
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <linux/if.h>
#include <linux/sockios.h>
#include <net/if_arp.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <stdlib.h>
#include <errno.h>

#include "bootp.h"

#define TAG_MERIT_DUMP    14
#define TAG_DOMAINNAME    15
#define TAG_SWAP_SERVER   16
#define TAG_ROOTPATH      17
#define TAG_EXTENSIONS    18

#define use_hwaddr ifr_hwaddr.sa_data

static void ParseCookie (unsigned char *cookie, unsigned char *match,
			 struct in_addr my_addr, unsigned long *netmask,
			 unsigned long *gateway);

static int subnet = 0;

int
do_bootp (const char *device,
	  unsigned long *ip, unsigned long *netmask, unsigned long *gateway)
{
    int sockfd, one = 1, i, retry_wait, waited;
    struct sockaddr_in cli_addr, serv_addr;
    struct bootp bppdu;
    struct ifreq ifr;
    fd_set rfds, wfds, xfds;
    struct timeval timeout;
    long rancopy;
    struct in_addr temp_addr;
    struct timeval tp;
    unsigned char mincookie[] =
    {99, 130, 83, 99, 255};

    memset (&serv_addr, 0, sizeof (serv_addr));

    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = -1;
    serv_addr.sin_port = htons (IPPORT_BOOTPS);

    if ((sockfd = socket (AF_INET, SOCK_DGRAM, 0)) < 0) {
	/* XXX perror ("bootpc: socket failed"); */
	return -1;
    }
    if (setsockopt (sockfd, SOL_SOCKET, SO_BROADCAST,
		    &one, sizeof (one)) == -1) {
	/* XXX perror ("bootpc: setsockopt failed"); */
	return -1;
    }
    memset (&cli_addr, 0, sizeof (cli_addr));
    cli_addr.sin_family = AF_INET;
    cli_addr.sin_addr.s_addr = htonl (INADDR_ANY);
    cli_addr.sin_port = htons (IPPORT_BOOTPC);

    if (bind (sockfd, (struct sockaddr *) &cli_addr, sizeof (cli_addr)) < 0) {
	/* perror ("bootpc: bind failed"); */
	return -1;
    }
    memset (&bppdu, 0, sizeof (bppdu));

    memcpy (ifr.ifr_name, device, strlen (device) + 1);
    if (ioctl (sockfd, SIOCGIFHWADDR, &ifr) < 0) {
	/* perror ("bootpc: ioctl failed"); */
	return -1;
    }
    bppdu.bp_htype = ifr.ifr_hwaddr.sa_family;
    bppdu.bp_hlen = IFHWADDRLEN;

    if (gettimeofday (&tp, NULL) == -1)
	rancopy = 0;
    else
	rancopy = tp.tv_sec + tp.tv_usec;

    for (i = 0; i < IFHWADDRLEN; ++i) {
	((unsigned char *) &rancopy)[i % sizeof (rancopy)] ^=
	    ((unsigned char *) (ifr.use_hwaddr))[i];
    }
    srand (rancopy);

    bppdu.bp_op = BOOTREQUEST;

    switch (bppdu.bp_htype) {
    case ARPHRD_ARCNET:
	memcpy (bppdu.bp_chaddr + IFHWADDRLEN - 1, ifr.use_hwaddr, 1);
	bppdu.bp_htype = ARPHRD_ETHER;
	bppdu.bp_hlen = IFHWADDRLEN;
	break;

    default:
	memcpy (bppdu.bp_chaddr, ifr.use_hwaddr, bppdu.bp_hlen);
    }

    bppdu.bp_secs = 0;

    memcpy (bppdu.bp_vend, mincookie, sizeof (mincookie));

    memcpy (bppdu.bp_file, "", 1);

    retry_wait = 2;
    while (retry_wait <= 64) {

	bppdu.bp_xid = rancopy = rand ();

	if (sendto (sockfd, &bppdu, sizeof (bppdu), 0,
		  (struct sockaddr *) &serv_addr, sizeof (serv_addr)) < 0) {
	    /* XXX perror ("bootpc: sendto"); */
	    return -1;
	}
	FD_SET (sockfd, &rfds);
	FD_ZERO (&wfds);
	FD_ZERO (&xfds);

	timeout.tv_sec = waited = retry_wait +
	    (1 + (rand () & (retry_wait - 1)));
	timeout.tv_usec = 0;

	if (select (sockfd + 1, &rfds, &wfds, &xfds, &timeout) < 0) {
	    /* perror ("bootpc: select"); */
	    return -1;
	}
	if (!FD_ISSET (sockfd, &rfds)) {
	    bppdu.bp_secs += waited;
	    retry_wait = retry_wait * 2;
	} else {
	    if (recvfrom (sockfd, &bppdu, sizeof (bppdu), 0,
			  (struct sockaddr *) NULL, (int *) NULL) < 0) {
		/* perror ("bootpc: recvfrom"); */
		return -1;
	    }
	    if (bppdu.bp_xid != rancopy) {
		/* fprintf(stderr, "WARNING bp_xid mismatch\n") ; */
		/* Should we quit here, ignore the packet or carry on??? */
	    }
	    *ip = bppdu.bp_yiaddr.s_addr;

	    ParseCookie ((unsigned char *) &bppdu.bp_vend, mincookie,
			 temp_addr, netmask, gateway);

	    if (!subnet) {
		int type = ntohl (temp_addr.s_addr);
		if ((type & 0x80000000) == 0) {		/* Class A */
		    *netmask = htonl (0xFF000000);
		} else if ((type & 0x40000000) == 0) {	/* Class B */
		    *netmask = htonl (0xFFFF0000);
		} else if ((type & 0x20000000) == 0) {	/* Class C */
		    *netmask = htonl (0xFFFFFF00);
		} else {	/* GOD KNOWS... other classes are weird */
		    *netmask = htonl (0x00000000);
		}
	    }
	    return 0;
	}
    }
    return -1;
}

static void
ParseCookie (unsigned char *cookie, unsigned char *match,
	     struct in_addr my_addr, unsigned long *netmask,
	     unsigned long *gateway)
{
    int i = 0, len;
    struct in_addr temp;

    for (i = 0; i < 4; ++i) {
	if (cookie[i] != match[i])
	    return;
    }

    while (i < 64) {
	switch (cookie[i]) {	/* The monster switch statement ... */
	case TAG_PAD:		/* PAD cookie */
	    i++;
	    break;
	case TAG_SUBNET_MASK:	/* SUBNET we are in */
	    memcpy (netmask, cookie + i + 2, 4);
	    subnet = 1;
	    i += 6;
	    break;
	case TAG_GATEWAY:	/* IP Gateways (routers) */
	    len = cookie[i + 1];
	    if (len >= 4) {
		*gateway = *(unsigned long *) (cookie + (i + 2));
	    }
	    i += len + 2;
	    break;
	case TAG_TIME_OFFSET:	/* Time of day */
	case TAG_TIME_SERVER:	/* Timeservers (see RFC-868) */
	case TAG_NAME_SERVER:	/* IEN-116 Nameservers */
	case TAG_DOMAIN_SERVER:	/* DNS Nameservers */
	case TAG_LOG_SERVER:	/* LOGGING servers */
	case TAG_COOKIE_SERVER:	/* Quote of day/Cookie servers */
	case TAG_LPR_SERVER:	/* LPR servers */
	case TAG_IMPRESS_SERVER:	/* Impress (Imogen) servers */
	case TAG_RLP_SERVER:	/* Remote Location Protocol servers */
	case TAG_BOOTSIZE:	/* BOOT File Size (ignored) */
	case TAG_MERIT_DUMP:	/* Merit DUMP File name (ignored) */
	case TAG_ROOTPATH:	/* Root pathname to mount as root filesystem */
	case TAG_EXTENSIONS:	/* Extensions.  Name of further Cookie data */
	case TAG_HOSTNAME:	/* HOSTNAME (may be fqdn or leaf) */
	case TAG_DOMAINNAME:	/* DOMAIN */
	    len = cookie[i + 1];
	    i += len + 2;
	    break;
	case TAG_SWAP_SERVER:	/* SWAPServer address */
	    memcpy (&temp, cookie + i + 2, 4);
	    i += 5;
	    break;
	case TAG_END:		/* END of cookie (phew) */
	    return;
	default:
	    if (cookie[i] >= 128 && cookie[i] <= 254) {
		i += 2 + cookie[i + 1];
	    }
	    ++i;
	    break;
	}
    }
}
