/* Copyright (C) 1995 Florian La Roche */
/* Who wants to help coding? I don't like doing this... */

/* You can just start setup as normal user and see how far it is coded
   right now. This will do a fake installation and won't actually chnage
   any data on your computer. */

/* Show an extra text-box with the contents of all external commands,
   before they are executed. So you can abort the installation, if any
   wrong commands are to be executed. (So don't format wrong partition.) */
/* #define VERBOSE 1 */

/* If defined, don't actually execute any comands and don't actually modify
   any files. So you can test any possible installation without doing any
   damage to your computer.
   The file FDISK.TEST is used instead of real "fdisk -l" output, so that
   it can be started as normal user. */
#if 0
#define DEBUG_THIS 1
#endif

#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <ctype.h>
#include <curses.h>
#include <stdarg.h>
#include <unistd.h>
#include <arpa/inet.h>
#include "dialog.h"

/* from bootpc.c */
int do_bootp(const char *device,
        unsigned long *ip, unsigned long *netmask, unsigned long *gateway);

/* max length of a partition name like e.g. '/dev/hda1' */
#define MAX_DEV_NAME 25

/* max number of possible Linux/Swap/MsDos partitions */
#define MAX_PARTS 20

char *progname = NULL;

void error (const char *s) {
	fprintf(stderr, "%s: %s\n", progname, s);
	exit(1);
}

int my_system (const char *s, ...) {
	int ret, i;
	va_list ap;
	char sh[8000];

	va_start(ap, s);
	vsprintf(sh, s, ap);
	va_end(ap);

#ifdef	VERBOSE
	i = dialog_msgbox("I will run the following command:", sh, 14, 65, 1);
	dialog_clear();
#endif
#ifdef DEBUG_THIS
	return 0;
#endif
	ret = system(sh);
	if (!(ret >> 8))
		return 0;
	i = dialog_msgbox("Error-Exit on the following command:",
		sh, 12, 73, 1);
	dialog_clear();
	return 1;
}

int have_ide_disks = 0;

/* We support to install from DOS/Linux-partitions. */
enum partition_type {
	MsDos,
	Linux,
	Swap
};

struct partition {
	enum partition_type type;
	char name[MAX_DEV_NAME];
	int blocks;
	int flag;
}partitions[MAX_PARTS];
int num_partition = 0;
int num_linux = 0;
int num_swap = 0;
int num_msdos = 0;

int get_line(char *line, int size, FILE *f) {
	char *ptr = line;
	int c;

	if (feof(f))
		return -1;
	while (size-- && ((c = getc(f)) != EOF) && (c != '\n'))
		*ptr++ = c;
	*ptr = '\0';
	return (int)(ptr - line);
}

int read_partitions (void) {
	FILE *f;
	char line[200];
	int length;
#ifndef DEBUG_THIS
	int ret = system("fdisk -l 2>/dev/null 1>/tmp/fdisk.output");
	if ((ret >> 8) != 0) {
		error("fdisk didn't run");
	}
	if ((f = fopen("/tmp/fdisk.output", "r")) == NULL)
#else
	if ((f = fopen("FDISK.TEST", "r")) == NULL)
#endif
		error("cannot read fdisk output");

	while (num_partition <= MAX_PARTS
			&& (length = get_line(line, 200, f)) >= 0) {
		if (strstr(line, "/dev/hda"))
			have_ide_disks = 1;
		if (strncmp(line, "/dev/", 5) == 0) {
			int n = 0;
			char *s = line + 5;
			char *t = partitions[num_partition].name;
			strcpy(t, "/dev/");
			t += 5;
			while (n < MAX_DEV_NAME && *s != '\0'
					&& !isspace((unsigned char)*s)) {
				*t++ = *s++;
				n++;
			}
			*t = '\0';
			/* Read the size of the partition. */
			t = line + 37;
			while (isspace((unsigned char)*t))
				t++;
			partitions[num_partition].blocks = atoi(t);
			if (strstr(line, "Linux native")) {
				partitions[num_partition].type = Linux;
				num_partition++;
				num_linux++;
			} else if (strstr(line, "Linux swap")) {
				partitions[num_partition].type = Swap;
				num_partition++;
				num_swap++;
			} else if (strstr(line, "DOS")) {
				partitions[num_partition].type = MsDos;
				num_partition++;
				num_msdos++;
			}
		}
	}
	fclose(f);
#ifndef DEBUG_THIS
	unlink("/tmp/fdisk.output");
#endif
	return 0;
}

int select_partition (const char *title, const char *prompt, int y, int x) {
	int i, num, ret;
	char info[MAX_PARTS][40];
	const char *items[MAX_PARTS * 2];
	int num_pa[MAX_PARTS];

	num = 0;
	for (i = 0; i < num_partition; i++) {
		if (partitions[i].type == Linux) {
			items[num*2] = partitions[i].name;
			sprintf(info[num], "Linux partition with %d blocks",
				partitions[i].blocks);
			items[num*2+1] = info[num];
			num_pa[num] = i;
			num++;
		}
	}
	ret = dialog_menu(title, prompt, y + num, x, num, num, items);
	dialog_clear();
	if (ret >= 0)		/* item selected */
		ret = num_pa[ret];
	return ret;
}

int select_install_partition (void) {
	return select_partition("Select Install Partition",
		"\\nWhere do you want to install Linux?\\n", 9, 60);
}

int select_source_partition (void) {
	return select_partition("Select Source Partition",
		"\\nOn which partition is the source?\\n", 9, 60);
}

const char *null = ">/dev/null 2>/dev/null";
const char *install_partition = NULL;

/* XXX Ugly restriction. */
#define MAX_TAR_FILES 200
struct file {
	char name[9];
	long size;
	long yes;
	char *des;
	char *help;
};
struct dir {
	char name[9];
	int yes;
	int num_files;
	char *des;
	struct file file[MAX_TAR_FILES];
};
#define MAX_NUM_DIRS 20
int num_dirs;
struct dir dirs[MAX_NUM_DIRS];

int read_contents (const char *source_path) {
	FILE *f;
	int length, n;
	struct dir *d = NULL;
	char s[200];
#ifndef	DEBUG_THIS
	sprintf(s, "/source%s/contents", source_path);
#else
	strcpy(s, "contents");
#endif
	if ((f = fopen(s, "r")) == NULL) {
		/* XXX int i = */ dialog_msgbox("Error", "Cannot open contents.",
			12, 60, 1);
		dialog_clear();
		return 1;
	}
	num_dirs = 0;
	while ((length = get_line(s, 200, f)) >= 0) {
		if (s[0] == '#')
			;	/* comment line */
		else if (strncmp(s, "DIR ", 4) == 0) {
			if (num_dirs >= MAX_NUM_DIRS) break;
			if (length < 7) break;
			if (s[5] != ' ') break;
			d = &dirs[num_dirs++];
			d->num_files = 0;
			d->yes = (s[4] == 'y')? 1 : 0;
			n = 0;
			while (n < 8 && isalnum((unsigned char)s[6 + n])) {
				d->name[n] = s[6 + n];
				n++;
			}
			d->name[n] = '\0';
			d->des = strdup(s + 7 + n);
		} else if (strncmp(s, "TAR ", 4) == 0) {
			struct file *t;
			if (d == NULL) break;
			if (d->num_files >= MAX_TAR_FILES) break;
			if (length < 9) break;
			if (s[5] != ' ') break;
			t = &d->file[d->num_files++];
			t->yes = 0;
			if (s[4] == 'y') t->yes = 1;
			if (s[4] == 'Y') t->yes = 2;
			n = 0;
			while (n < 8 && isalnum((unsigned char)s[6 + n])) {
				t->name[n] = s[6 + n];
				n++;
			}
			t->name[n] = '\0';
			n += 7;
			t->size = atol(s + n);
			while (isdigit((unsigned char)s[n])) n++;
			t->des = strdup(s + n + 1);
		} else if (strncmp(s, "DES ", 4) == 0) {
		} else if (length == 0) {
		} else {
			/* XXX int i = */ dialog_msgbox("Error - Wrong line",
				s, 12, 60, 1);
			dialog_clear();
			/* XXX */
			continue;
			fclose(f);
			return 1;
		}
	}
	fclose(f);
	return 0;
}

#define MAX(a,b) (((a) > (b))?(a):(b))
#define MAX_SELECTION MAX(MAX_NUM_DIRS,MAX_TAR_FILES)

int select_tar_files(void) {
	int ret, n, i;
	const char *d[MAX_SELECTION * 3];
	int status[MAX_SELECTION];

	for (n = 0; n < num_dirs; n++) {
		d[n*3] = dirs[n].name;
		d[n*3 + 1] = dirs[n].des;
		d[n*3 + 2] = dirs[n].yes? "on" : "off";
	}
	ret = dialog_checklist ("Select Packages to Install",
		"Choose the main packages to install onto your hd.",
		8 + num_dirs, 70, num_dirs, num_dirs, d, 1, status);
	dialog_clear();
	for (n = 0; n < num_dirs; n++)
		dirs[n].yes = (status[n] == 1)? 1 : 0;
	for (n = 0; n < num_dirs; n++) {
		struct dir *f = &dirs[n];
		int l;
		if (f->yes == 0) continue;
		for (i = 0; i < f->num_files; i++) {
			if (f->file[i].yes == 2) break;
			d[i*3] = f->file[i].name;
			d[i*3 + 1] = f->file[i].des;
			d[i*3 + 2] = (f->file[i].yes? "on" : "off");
		}
		l = (i > 14)? 14 : i;
		ret = dialog_checklist ("Select Packages to Install",
			"Choose the packages to install onto your hd.",
			8 + l, 70, l, i, d, 1, status);
		dialog_clear();
		for (l = 0; l < i; l++)
			f->file[l].yes = (status[l] == 1)? 1 : 0;
	}

	return 0;
}

int extract_packages (const char *source_path) {
	int ret, i, j;
	char *mbr_part;		/* name of the master boot record MBR */
#ifndef	DEBUG_THIS
	FILE *f;
#endif
	char * const lilo_part[] = {
		"1", "beginning of the Linux partition",
		"2", "beginning of the harddisk /dev/hda",
		"3", "don't install it" };

	ret = my_system("mkdir -p /install/var/installed/packages %s", null);
	if (ret) return ret;
	ret = read_contents(source_path);
	if (ret) {
		/* XXX int i = */ dialog_msgbox("Error",
			"Error reading contents.", 12, 60, 1);
		dialog_clear();
		return ret;
	}
	ret = select_tar_files();
	if (ret) return ret;
	for (i = 0; i < num_dirs; i++) {
		struct dir *d = &dirs[i];
		char msg[] = "\n  Please be patient."
			     "\n  I install files from                ";
		char *s, tars[MAX_TAR_FILES * 9 + 400];
		if (! d->yes) continue;

		strcpy(msg + 45, d->name);
		strcpy(msg + strlen(msg), ".");
		ret = dialog_msgbox("Installing...", msg, 10, 65, 0);

		strcpy(tars, "cd /install; for i in");
		s = tars + strlen(tars);
		for (j = 0; j < d->num_files; j++) {
			struct file *f = &d->file[j];
			if (!f->yes) continue;
			*s++ = ' ';
			strcpy(s, f->name); s += strlen(s);
		}
		sprintf(s, "; do tar xzplvvkf /source%s/%s/$i.tgz "
			"> var/installed/packages/$i "
			"2>> var/installed/packages/ERROR; done",
			source_path, d->name);
		ret = my_system(tars);
		dialog_clear();
		/* if (ret) return ret; */
	}
	mbr_part = "/dev/hda";
	if (! have_ide_disks) {
		mbr_part = "/dev/sda";
		lilo_part[3][31] = 's';
	}
	ret = dialog_menu ("Install Linux Loader LILO",
		"\nI suggest installing LILO at the first sector of your\n"
		"root partition. Use fdisk to make that partition\n"
		"bootable and add entries for your DOS partitions there.\n"
		"Only use the MBR, if you have Linux on your 2nd hd.\n\n"
		"Where should I install LILO ?\n\n", 19, 64, 3, 3, lilo_part);
	dialog_clear();
	if (ret == 0 || ret == 1) {
#ifndef	DEBUG_THIS
		if ((f = fopen("/install/etc/lilo.conf", "w")) == NULL) {
			i = dialog_msgbox("Error",
				"Cannot write /etc/lilo.conf", 12, 40, 1);
			dialog_clear();
			return -1;
		}
		fprintf(f, "#append=\"...\"\n");
		fprintf(f, "boot=%s # where lilo is installed\n",
			(ret == 0)? install_partition : mbr_part);
		fprintf(f, "delay=30 # 3 secs before starting default\n");
		fprintf(f, "vga=normal # I use vga=1\n");
		fprintf(f, "ramdisk=0\n");
		fprintf(f, "root=%s # where linux is installed\n",
			install_partition);
		fprintf(f, "image=/boot/zImage\n");
		fprintf(f, "  label=linux\n");
		fprintf(f, "  read-only\n");
		fclose(f);
#endif
		ret = my_system("lilo -r /mnt %s", null);
		/* if (ret) return ret;	*/
	}

#ifndef	DEBUG_THIS
	if ((f = fopen("/install/etc/fstab", "w")) == NULL) {
		i = dialog_msgbox("Error", "Cannot write /etc/fstab",
			12, 40, 1);
		dialog_clear();
		return -1;
	}
	fprintf(f, "%s / ext2 defaults 1 1\n", install_partition);
	fprintf(f, "none /proc proc defaults 0 2\n");
	/* XXX write swap-partitions */
	fclose(f);
#endif
	ret = dialog_msgbox("Installation complete.",
		"\n  Installation is now comlete.\n",
		6, 40, 1);
	dialog_clear();
	return 0;
}

int install_premounted (void) {
	return extract_packages("");
}

int install_harddisk (void) {
	const char *name;
	int part, ret;

	if ((part = select_source_partition()) <= -1) return part;
	name = partitions[part].name;

	ret = my_system("mount -t ext2 %s /source %s", name, null);
	if (ret) return ret;
	ret = dialog_inputbox("Path in partition",
		"Please enter the directory in which the "
		"source files are.", 13, 50, "");
	dialog_clear();
	if (ret) return ret;
			/* XXX strdup */
	ret = extract_packages(strdup(dialog_input_result));
	if (ret) return ret;
	ret = my_system("umount /source %s", null);
	if (ret) return ret;
	return 0;
}

#define CONVERT_NTOA(y) x.s_addr = y; strcpy(y ## _a, inet_ntoa(x))

int install_nfs (void) {
	const char *eth_types[] = { "Internal", "Internal ethernet card",
		"External", "External D-Link adaptor" };
	const char * const nfs_yesno[] = {
		"NFS", "be asked about all network configurations",
		"NFS+bootp", "use bootp to get most net configs",
		"no", "do not install via NFS, return to last menue" };
	unsigned long ip, netmask, network, broadcast, gateway, server;
	char ip_a[20], netmask_a[20], network_a[20], broadcast_a[20],
		gateway_a[20], server_a[20];
	struct in_addr x;
	char *path, *type;
	int bootp = 0, ret;

	ret = dialog_menu ("Installation Through NFS",
		"\nIf you want to install via network, you will need "
		"some other\ncomputer as server for all files."
		"\n\nHere is a list of things, you need to know:\n"
		"- Your own IP address and your netmask.\n"
		"- The gateway of your local net, if the server is not "
		"in it.\n"
		"- The IP address of the server and the proper source "
		"directory.\n", 18, 74, 3, 3, nfs_yesno);
	dialog_clear ();
	if (ret < 0 || ret == 2) return ret;
	if (ret == 1) bootp = 1;

	ret = dialog_menu ("Ethernet Device Selection",
		"\nWhich type of ethernet device do you have?\n \n",
		12, 65, 2, 2, eth_types);
	dialog_clear ();
	if (ret) return ret;
	type = (ret == 0)? "eth0" : "dl0";

	if (bootp) {
		ret = my_system("ifconfig %s up %s", type, null);
		ret = my_system("route add default dev %s %s", type, null);
		ip = netmask = gateway = 0;
		ret = do_bootp(type, &ip, &netmask, &gateway);
		if (ret) {
			bootp = 0;
			ip_a[0] = '\0';
		} else {
			network = ip & netmask;
			broadcast = network | ~netmask;
			CONVERT_NTOA(ip);
		}
		ret = my_system("route del default %s", null);
		ret = my_system("ifconfig %s down %s", type, null);
	} else ip_a[0] = '\0';
	ret = dialog_inputbox ("Assign IP Address",
		"\nYou will need to enter the IP address you wish to\n"
		"assign to this machine. Example: 111.112.113.114\n\n"
		"What is your IP address ?", 13, 58, ip_a);
	dialog_clear ();
	if (ret) return ret;
	ip = inet_addr (dialog_input_result);

	if (bootp) {
		CONVERT_NTOA(netmask);
	} else strcpy(netmask_a, "255.255.255.0");
	ret = dialog_inputbox ("Assign Netmask",
		"\nNow we need to know your netmask. Typically this\n"
		"will be 255.255.255.0, but this can be different\n"
		"depending on your local setup.\n\n"
		"What is your netmask ?", 14, 56, netmask_a);
	dialog_clear ();
	if (ret) return ret;
	netmask = inet_addr (dialog_input_result);

	network = ip & netmask;
	broadcast = network | ~netmask;
	if (!bootp)
		gateway = network | 0x01000000ul;

	ret = dialog_yesno ("", "\nDo you have a gateway ?", 7, 30);
	dialog_clear ();
	if (ret == -1) return ret;
	if (ret == 0) {
		CONVERT_NTOA(gateway);
		ret = dialog_inputbox ("Assign Gateway IP Address",
			"\nWhat is your gateway IP address?\n",
			10, 50, gateway_a);
		dialog_clear ();
		if (ret) return ret;
		gateway = inet_addr (dialog_input_result);
	} else
		gateway = -1;

	ret = dialog_inputbox ("IP Address of NFS Server",
		"\nWhat is the IP address of your NFS server ?\n"
		"(Remote computer giving you the needed files.)\n \n",
		12, 65, "");
	dialog_clear ();
	if (ret) return ret;
	server = inet_addr (dialog_input_result);

	ret = dialog_inputbox ("Select Source Directory",
		"\nNow we need to know where the rest of the "
		"distribution\ntar files can be found. Normally, this will "
		"be in a\nfurther subdir called \"tar\" from the main "
		"directory.\n(E.g. something like /linux/distribution/tar.)\n"
		"\nWhat directory are further distribution files in?\n",
		15, 65, "");
	dialog_clear ();
	if (ret) return ret;
	path = strdup (dialog_input_result);
	/* XXX path == NULL ?? give better warning */
	if (path == NULL) return -1;

	CONVERT_NTOA(ip);
	CONVERT_NTOA(netmask);
	CONVERT_NTOA(network);
	CONVERT_NTOA(broadcast);
	CONVERT_NTOA(gateway);
	CONVERT_NTOA(server);

	ret = dialog_msgbox("Please Wait",
		"\n  I will set up the network devices and try to make\n"
		  "  a connection to your NFS server. This can take\n"
		  "  some time to finish...\n\n"
		  "  Please wait.\n", 9, 63, 0);
#ifdef  DEBUG_THIS
	sleep(3);
#endif
	if (ret) return ret;
 
	ret = my_system("ifconfig %s %s netmask %s broadcast %s %s",
		type, ip_a, netmask_a, broadcast_a, null);
	if (ret) return ret;
	ret = my_system("route add -net %s %s", network_a, null);
	if (ret) return ret;
	if (gateway != -1) {
		ret = my_system("route add default gw %s metric 1 %s",
			gateway_a, null);
		if (ret) return ret;
	}
	ret = my_system("mount -t nfs %s:%s /source %s", server_a, path, null);
	if (ret) return ret;
	ret = extract_packages("");
	if (ret) return ret;
	ret = my_system("umount /source %s", null);
	if (ret) return ret;
	ret = my_system("ifconfig eth0 down %s", null);
	if (ret) return ret;
	return 0;
}

void main_install (void) {
	int part, ret;
	const char *name;
	const char *items1 [] = {
		"1",	"Harddisk Install",
		"2",	"Network Install(NFS)",
		"3",	"Premounted on /source"
	};

	if (num_linux == 0) {
		/* XXX */
		return;
	}
	if ((part = select_install_partition()) <= -1) return;
	install_partition = name = partitions[part].name;
	if (my_system("mke2fs %s %s", name, null)) return;
	if (my_system("mount -t ext2 %s /install %s", name, null)) return;
	ret = dialog_menu("Choose install medium",
		"\\nPlease say from where you want to install.\\n",
		12, 62, 3, 3, items1);
	dialog_clear();
	switch (ret) {
	    case 0:	ret = install_harddisk(); break;
	    case 1:	ret = install_nfs(); break;
	    case 2:	ret = install_premounted(); break;
	    case -2:			/* cancel */
	    case -1:	break;		/* esc */
	}
	if (my_system("umount /install %s", null))
		return;
}

int main (int argc, char **argv) {
	int stop = 0;
	int ret;
	const char *items1 [] = {
		"1",		"Display a help text",
		"2",		"Start an installation",
		"3",		"Exit to the shell"
	};

	progname = argv[0];

#if 0
	{ struct in_addr x;
	int ip = inet_addr ("134.96.116.11\n");
	x.s_addr = ip;
	printf(inet_ntoa(x));
	exit(1);
	}
#endif
	ret = read_partitions();
	if (num_linux == 0) {
		printf(	"\n\nPlease start \"fdisk\" or \"cfdisk\" and create a"
			"\nnative Linux-partition to install Linux on.\n\n");
		exit(1);
	}

	dialog_init();
#if 0
	{ int n;
	read_contents("");
	dialog_end();
	for (ret = 0; ret < num_dirs; ret++) {
		printf("\n%s %s", dirs[ret].name, dirs[ret].des);
		for (n = 0; n < dirs[ret].num_files; n++) {
			printf("\n%s %s", dirs[ret].file[n].name,
				dirs[ret].file[n].des);
		}
	}
	}
#else

	while (!stop) {
		ret = dialog_menu("Linux Install Utility",
			"\\nCopyright (C) 1995 Florian La Roche\\n"
			"\\nPre-Alpha version, be careful, read the doc!!!"
			"\\nemail: florian@jurix.jura.uni-sb.de, "
			"flla@stud.uni-sb.de\\n",
			15, 64, 3, 3, items1);
		dialog_clear();
		switch (ret) {
		    case 0:	ret = dialog_textbox("Help Text",
					"/etc/setup.help", 20, 70);
				dialog_clear();
				break;
		    case 1:	main_install(); break;
		    case 2:	stop = 1; break;
		    case -2:	/* cancel */
		    case -1:	stop = 1;	/* esc */
		}
	}
	dialog_end();
#endif
	printf("\nExecute \"reboot\" to restart your computer...\n");

	exit(0);
}

